/* ----------------------------------------------------
 *  Filename:           fmx.c
 *  Summary:            File Manager extension DLL
 *  Author:             T.W. Nelson
 *  Compiler:           BCC++ 4.0
 *  Compile options:    DLL, large model, C++ compile
 *  Start date:         11-Apr-1994
 *  Last update:        29-Jun-1994
 *  Version:            1.00
 *  Notes:
 *
 *  Source code Copyright (c) 1994 T.W. Nelson.
 *  May be used in any manner provided copyright
 *  notice is displayed appropriately.
 * ------------------------------------------------- */

#include <wfext.h>     /* includes windows.h */
#include <commdlg.h>
#include <string.h>    /* memset()  */
#include "fmx.h"

extern "C"  {   /* descramble for C++ compile.... */
LONG WINAPI FMExtensionProc(HWND, WORD, LONG);
LRESULT CALLBACK FMgetMsgHook(int,WORD,LONG);
LRESULT CALLBACK FMmouseMsgHook( int,WORD,LONG);
LRESULT CALLBACK KeyProcHook( int,WORD,LONG );
LRESULT CALLBACK MenuWndProc(HWND,UINT,UINT,LONG);
LRESULT CALLBACK FMmenuHook(int,WORD,LONG);
UINT CALLBACK ExecProgHook(HWND,UINT,WPARAM,LPARAM);
}

/* Run-time flags, with defaults as shown.  Can be
 * modified in WINFILE.INI if desired....
 */
int   fEnableFloat   = 1;  /* enable float menu */
int   fPrintID       = 0;  /* print menu IDs */
int   fRestrictFloat = 0;  /* flt menu in FM only */

/* Run-time data, most initialized in FMEVENT_LOAD. */
WORD      wDelta;          /* menu delta-factor */
POINT     Pt;              /* mouse pos for menu */
HINSTANCE hFMinst;         /* FM instance handle */
HINSTANCE hEXinst;         /* Extension DLL inst */
HTASK     hFMtask;         /* FM task handle */
HHOOK     hMouseHook;      /* Hook proc handles */
HHOOK     hMsgHook;
HHOOK     hKeyHook;
HHOOK     hMenuHook;
HWND      hFMwnd;          /* file manager window */
HWND      hWndPopup;       /* popup menu window */
HWND      hExecDlg = 0;    /* commdlg hwnd */
HMENU     hFMmenu;         /* FM's main menu */
HMENU     hExMenu;         /* our extension menu */
int       fMenuActive = 0; /* FM menu active */
COMMANDTABLE *pCmdTab = NULL;
FMS_GETFILESEL  file;
char szFilter[256];
char szMenuTitle[20];

#define EXECFLOAT       FMMENU_LAST
#define INI_SECTION_HDR "ExtensionMenu"
#define lstrequ(a,b)     !lstrcmp(a,b)

COMMANDTABLE   cmdTab[] = {
/* These items define the extension menu ...... */
-1,-1,"DOS &Box",           -1,0,     WINEXEC,"dosprmpt.pif",
-1,-1,"&Calculator",        -1,0,     WINEXEC,"calc",
-1,-1,"Wi&nsight",          -1,0,     WINEXEC,"winsight",
-1,-1,"&DOS Debugger",      -1,0,     BINDEXEC,"td ",
-1,-1,"&Windows Debugger",  -1,0,     BINDEXEC,"tdw ",
-1,-1,"Note&pad\tF9",       -1,VK_F9, BINDEXEC,"notepad ",
-1,-1,"&Run...",            -1,0,     CMDEXEC, "",
-1,-1,"SEPARATOR",          -1,0,     NULLCMD, "",
-1,-1,"&Screen Saver\tF12", -1,VK_F12,SCREENSAVE, "",
-1,-1,"Reload Extensions",  -1,0,     EXTRELOAD, "",
-1,-1,"Restore File Manager",-1,0,    FMRESTORE, "",
-1,-1,"SEPARATOR",          -1,0,     NULLCMD, "",
-1,-1,"About...",           -1,0,     ABOUT, "",

/* These items modify action of existing FM menu
 * items in submenu 0 ...... */
 0,-1,"Search...\tF4",     104,VK_F4, NULLCMD, "",
 0,-1,"Select...\tF2",     116,VK_F2, NULLCMD, "",
 0,-1,"&Run...  \tF3",     105,VK_F3, CMDEXEC, "",

/* These items add new items to FM's "Help" popup
   menu, submenu #6 ...... */
 6, 3,"Windows API Help",   -1,0,     WINHELP,"bwinapi.hlp",
 6, 3,"Borland C++ Help",   -1,0,     WINHELP,"bcw.hlp",
    };

#define CMDTABSIZE ( sizeof(cmdTab)/ \
                     sizeof(COMMANDTABLE) )

/* ------------------------------------------------- */
int CALLBACK LibMain( HINSTANCE hinst, WORD wSeg,
                      WORD wHeap, LPSTR cmdline )
{
    hEXinst = hinst;
    return 1;
}

int CALLBACK WEP( int exitType )
{
    MessageBeep(0);  /* confirms DLL unload */
    return 1;
}

LPSTR WINAPI
GetSetupString( LPCSTR szEntry, LPCSTR szDef )
{
    static char retBuf[200];

    GetPrivateProfileString( INI_SECTION_HDR,
            szEntry, szDef, (LPSTR) retBuf,
            sizeof(retBuf), "winfile.ini" );

    return (LPSTR) retBuf;
}

UINT WINAPI
GetSetupInt( LPCSTR szEntry, int ndefault )
{
    return GetPrivateProfileInt( INI_SECTION_HDR,
            szEntry, ndefault, "winfile.ini" );
}

void WINAPI GetWinfileArgs(void)
{
    int i;
    char ch;    /* separator char */

    fEnableFloat = GetSetupInt( "EnableFloat",
                                 fEnableFloat  );
    fPrintID = GetSetupInt( "PrintID",
                             fPrintID   );
    fRestrictFloat = GetSetupInt( "RestrictFloat",
                                   fRestrictFloat  );
    lstrcpy( szMenuTitle,
             GetSetupString( "MenuTitle",
                             "E&xTools" ) );

    /* Get filter string for ExecProgram()...... */
    lstrcpy( szFilter, GetSetupString( "Filter",
                           ".EXE files|*.exe|" ) );
    ch = szFilter[ lstrlen(szFilter) - 1 ];

    /* Replace separator chars with \0 bytes. */
    for( i = 0; szFilter[i]; i++ )
        if( szFilter[i] == ch )
            szFilter[i] = '\0';
}

HMENU WINAPI SetupExtensionMenu( void )
{
    int i;
    HMENU hmenu;
    COMMANDTABLE *pcmd;

    hmenu = CreatePopupMenu();

    for( i = 0; i < CMDTABSIZE; i++ )   {
        pcmd = &cmdTab[i];

        if( pcmd->nExtCmd != NULLCMD )
            pcmd->nExtCmd += wDelta;

        if( pcmd->nMenuCmd == -1 )
            pcmd->nMenuCmd = wDelta + i + 1;

        if( lstrequ( pcmd->szMenuTxt, "SEPARATOR" ) )
            AppendMenu( hmenu, MF_SEPARATOR, 0, 0 );
        else if( pcmd->nPopup == -1 )   {
            /* Add to extension menu.... */
            AppendMenu( hmenu, MF_ENABLED,
                    i + 1, pcmd->szMenuTxt );
        }
        else if( pcmd->nItemPos == -1 )     {
            /* modify FM menu item by Command */
            ModifyMenu( GetSubMenu( hFMmenu,
                    pcmd->nPopup), pcmd->nMenuCmd,
                    MF_ENABLED | MF_BYCOMMAND,
                    pcmd->nMenuCmd, pcmd->szMenuTxt );
        }
        else {
            /* insert new item by Position
             * in existing FM menu .... */
            InsertMenu( GetSubMenu( hFMmenu,
                    pcmd->nPopup), pcmd->nItemPos,
                    MF_ENABLED | MF_BYPOSITION,
                    pcmd->nMenuCmd, pcmd->szMenuTxt );
        }
    }

    return hmenu;
}

LRESULT CALLBACK _export
MenuWndProc( HWND hwnd, UINT msg,
             UINT wParam, LONG lParam )
{
   /* Transfer WM_COMMAND messages from the
    * floating popup menu to FM's Window Proc.  These
    * messages are seen first by the FMgetMsgHook()
    * function, before they get to FM's WndProc().
    */

    if( msg == WM_COMMAND )     {
        SendMessage( hFMwnd, msg, wParam, lParam );
        return 0;
    }

    return DefWindowProc( hwnd, msg, wParam, lParam);
}

HWND WINAPI CreatePopupMenuWindow(void)
{
   /* Create an invisible window over which the
    * floating popup menu is activated.  This gives
    * our menu the keybd focus within any app.
    */

    static char classname[] = "FM_FloatMenu";
    WNDCLASS wc;
    memset( (void *) &wc, '\0', sizeof(WNDCLASS) );
    wc.lpfnWndProc   = MenuWndProc;
    wc.hInstance     = hEXinst;
    wc.lpszClassName = classname;
    RegisterClass( &wc );
    return CreateWindow( classname, "", NULL,
                            0, 0, 0, 0, NULL, NULL,
                            hEXinst, NULL );
}

VOID WINAPI PopupFloatMenu( void )
{
   /* Activate the floating popup menu.  Menu gets
    * keybd control by setting focus to the invisible
    * window prior to calling TrackPopupMenu().
    */

    HWND hwnd;

    if( fRestrictFloat )
        ClientToScreen( hFMwnd, (LPPOINT) &Pt );
    else    {
        hwnd = GetFocus();
        SetFocus( hWndPopup );
    }

    TrackPopupMenu( hExMenu,
        TPM_LEFTALIGN | TPM_LEFTBUTTON, Pt.x, Pt.y, 0,
        fRestrictFloat ? hFMwnd : hWndPopup, NULL );

    if( !fRestrictFloat )
        SetFocus( hwnd );
}

LRESULT CALLBACK  _export  /* WH_MOUSE type */
FMmouseMsgHook( int nCode, WORD wParam, LONG lParam )
{
   /* Look for a WM_RBUTTONDOWN message and post
    * an EXECFLOAT message to FM when one is found.
    * If fRestrictFloat is non-zero, code restricts
    * float menu to FM client area only.
    */

    RECT rc;
    MOUSEHOOKSTRUCT *pmh;

    if( nCode < 0 )   /* don't process */
        return CallNextHookEx( hMouseHook, nCode,
                                  wParam, lParam );

    switch( wParam )
    {
    case WM_RBUTTONDOWN:
        if( !fEnableFloat || fMenuActive )
            break;  /* disabled, allow pass-thru */

        pmh = (MOUSEHOOKSTRUCT *) lParam;

        if( fRestrictFloat )    {
            /* This activates float menu only within
             * FM client area ....
             */
            GetClientRect( hFMwnd, (LPRECT) &rc );

            if( PtInRect( (LPRECT) &rc, pmh->pt ) ) {
                Pt = pmh->pt;   /* save mouse pos */
                PostMessage( hFMwnd, WM_COMMAND,
                             EXECFLOAT+wDelta, 0L );

                return 1L;  /* eat the message */
            }

            break;  /* pass thru */
        }
        else    {
           /* This does it anywhere ......... */
            Pt = pmh->pt;   /* save mouse pos */
            PostMessage( hFMwnd, WM_COMMAND,
                           EXECFLOAT+wDelta, 0L );

            return 1L;  /* eat the message */
        }

    default:
        break;
    }

    return 0L;  /* continue */
}

LRESULT CALLBACK _export    /* WH_MSGFILTER type */
FMmenuHook( int nCode, WORD wParam, LONG lParam )
{
   /* Set the state of the fMenuActive flag, which
    * will be non-zero if an FM menu is active, 0
    * otherwise.  This flag disables keybd accelerators
    * when FM is in menu mode.
    */

    LPMSG pm = (LPMSG) lParam;

    if( nCode < 0 )   /* don't process */
        return CallNextHookEx( hMenuHook, nCode,
                                 wParam, lParam );

    if( pm->message == WM_MENUSELECT )
        fMenuActive = pm->wParam;

    return 0L;   /* continue default processing */
}

LRESULT CALLBACK _export  /* WH_GETMESSAGE type */
FMgetMsgHook( int nCode, WORD wParam, LONG lParam )
{
    int i;
    char msgBuf[100];
    LPMSG pm = (LPMSG) lParam;

    if( nCode < 0 )   /* don't process */
        return CallNextHookEx( hMsgHook, nCode,
                                 wParam, lParam );

    switch( pm->message )
    {
    case WM_KEYDOWN:

       /* Match keypress with an accel key in cmdTab[].
        * If found, post the corresponding WM_COMMAND
        * menu item from cmdTab[] to FM's window.  This
        * code doesn't check for shift, Alt, or Cntrl
        * keys.
        */

        if( fMenuActive )    /* disable accelerators */
            break;

        for( i = 0; i < CMDTABSIZE; i++ )   {
            if( pm->wParam == cmdTab[i].nAccelKey ) {
                PostMessage( hFMwnd, WM_COMMAND,
                             cmdTab[i].nMenuCmd, 0L );
                break;
            }
        }

        break;

    case WM_COMMAND:
       /* Print menu ID (in wParam) for FM menu
        * commands........
        */

        if( fPrintID )    {
            char szMenuItem[80];
            GetMenuString( hFMmenu,
                pm->wParam, szMenuItem,
                sizeof(szMenuItem), MF_BYCOMMAND );
            wsprintf( msgBuf,
                  "wDelta = %d\n"
                  "Command ID = %u\nMenu item: %s",
                   wDelta, pm->wParam, szMenuItem );
            MessageBox( hFMwnd, msgBuf,
                "WM_COMMAND Menu Command", MB_OK );
        }

       /* Find a menu item & change to the
        * corresponding extension command......
        */

        for( i = 0; i < CMDTABSIZE; i++ )   {
            if( pm->wParam == cmdTab[i].nMenuCmd &&
                    cmdTab[i].nExtCmd != NULLCMD )  {
                pm->wParam = cmdTab[i].nExtCmd;
                pCmdTab = &cmdTab[i];
                break;
            }
        }

        break;

    default:
        break;
    }

    return 0L;  /* this hook always returns 0 */
}

LRESULT CALLBACK _export  /* WH_KEYBOARD type */
KeyProcHook( int nCode, WORD wParam, LONG lParam )
{
   /* This activates the popup menu in response to
    * an Alt+Insert keystroke, and is one way to avoid
    * possible conflicts with the right mouse button.
    * Somewhat crude, but it works.  NOTE: does NOT
    * work in a DOS box.....
    */

    if( nCode < 0 )   /* don't process */
        return CallNextHookEx( hKeyHook, nCode,
                                 wParam, lParam );

    if( !fRestrictFloat &&
         wParam == VK_INSERT &&
        (HIWORD(lParam) & KF_ALTDOWN) &&
       !(HIWORD(lParam) & KF_UP)         )    {
            Pt.x = GetSystemMetrics(SM_CXSCREEN)/2;
            Pt.y = GetSystemMetrics(SM_CYSCREEN)/2;
            PostMessage( hFMwnd, WM_COMMAND,
                         EXECFLOAT+wDelta, 0L );
            return 1;   /* eat it */
    }

    return 0;   /* continue */
}

UINT CALLBACK _export
ExecProgHook( HWND hdlg, UINT msg,
              WPARAM wParam, LPARAM lParam )
{
   /* Track the state of 'hExecDlg' handle for
    * the "ExecProgram" common dialog box.  The
    * handle should be non-0 only when the dialog box
    * is active, 0 otherwise.
    */

    switch( msg )   {
    case WM_INITDIALOG:
        hExecDlg = hdlg;    /* popped up */

        return TRUE;    /* end processing */

    case WM_COMMAND:
        if( wParam == IDCANCEL ||
            wParam == IDOK     ||
            wParam == IDABORT     )    {
            hExecDlg = 0;   /* popped down */
        }

        break;
    }

    return FALSE;   /* continue processing */
}

void WINAPI
ExecProgram( LPCSTR szAppName )
{
   /* Implement an application launcher using the
    * "open file" common dialog box.  Although
    * usually modal, the dialog box behaves
    * modelessly here.
    */

    int rval;
    char msgBuf[100];
    static OPENFILENAME ofn;
    static char szFile[256];
    static char szFileTitle[256];

    if( hExecDlg )  {
        /* Disallow multiple invocations */
        DestroyWindow( hExecDlg );
        hExecDlg = 0;
    }

    lstrcpy( szFile, szAppName );
    memset( (void *) &ofn, '\0',
            sizeof(OPENFILENAME) );
    ofn.lStructSize = sizeof(OPENFILENAME);
    ofn.hwndOwner = NULL;
    ofn.lpstrFilter = szFilter;
    ofn.nFilterIndex = 1;
    ofn.lpstrFile = szFile;
    ofn.nMaxFile = sizeof(szFile);

    ofn.lpstrFileTitle = szFileTitle;
    ofn.nMaxFileTitle = sizeof(szFileTitle);
    ofn.lpstrInitialDir = NULL;
    ofn.lpstrTitle = "Exec Program";
    ofn.lpfnHook = ExecProgHook;
    ofn.Flags = OFN_NOVALIDATE | OFN_HIDEREADONLY |
                OFN_ENABLEHOOK | OFN_NOCHANGEDIR  ;

    if( GetOpenFileName( &ofn ) )   {
        if( (rval = WinExec( ofn.lpstrFile,
                         SW_SHOWNORMAL )) <= 31 )  {
            wsprintf( msgBuf,
                      "Return value = %u", rval );
            MessageBox( hFMwnd, msgBuf,
                    "WinExec() Error", MB_OK );
        }
    }
}

VOID WINAPI
GetCurrentFileSel( LPFMS_GETFILESEL file )
{
   /* Get the current file selection in FM's
    * Directory or Search window, NULL if no
    * file is selected........
    */

    UINT rval;

    memset( (void *) file, '\0',
                    sizeof(FMS_GETFILESEL) );
    rval = SendMessage( hFMwnd, FM_GETFOCUS, 0, 0L );

    if( rval == FMFOCUS_DIR ||
        rval == FMFOCUS_SEARCH )    {
        SendMessage( hFMwnd, FM_GETFILESEL, 0,
               (LONG) (LPFMS_GETFILESEL) file );
    }
}

LONG WINAPI _export
FMExtensionProc( HWND hwnd, WORD wMsg, LONG lParam )
{
    switch( wMsg )  {
    case FMEVENT_LOAD:
       /* DLL is unloaded if an FMS_LOAD structure is
        * not initialized here.  This case must return
        * TRUE to realize a successful load .........
        */
        wDelta = ((LPFMS_LOAD) lParam)->wMenuDelta;
        ((LPFMS_LOAD) lParam)->dwSize =
                                 sizeof(FMS_LOAD);
        GetWinfileArgs();
        lstrcpy(((LPFMS_LOAD) lParam)->szMenuName,
                                       szMenuTitle );
        hFMmenu = GetMenu( hwnd );
        hExMenu = SetupExtensionMenu();
        ((LPFMS_LOAD) lParam)->hMenu = hExMenu;
        hFMwnd  = hwnd;
        hFMinst = GetModuleHandle( "winfile.exe" );
        hFMtask = GetWindowTask( hwnd );

       /* NOTE: See SDK Annotation #4 in W/DDJ
        * July 1994, pg. 73 regarding use of instance
        * handles for SetWindowsHookEx()
        */
        hMsgHook = SetWindowsHookEx( WH_GETMESSAGE,
                           (HOOKPROC) FMgetMsgHook,
                                   hFMinst, hFMtask );

        hMouseHook = SetWindowsHookEx( WH_MOUSE,
             (HOOKPROC) FMmouseMsgHook, hFMinst,
                 fRestrictFloat ? hFMtask : NULL );

        hKeyHook = SetWindowsHookEx( WH_KEYBOARD,
             (HOOKPROC) KeyProcHook, hFMinst, NULL );

        hMenuHook = SetWindowsHookEx( WH_MSGFILTER,
                           (HOOKPROC) FMmenuHook,
                                   hFMinst, hFMtask );

        hWndPopup = CreatePopupMenuWindow();

        return (LONG) TRUE;

    case FMEVENT_UNLOAD:
        UnhookWindowsHookEx( hMouseHook );
        UnhookWindowsHookEx( hMsgHook );
        UnhookWindowsHookEx( hKeyHook );
        UnhookWindowsHookEx( hMenuHook );
        DestroyWindow( hWndPopup );
        DestroyMenu( hExMenu );

        /* Cancel dialog box if still active before
         * FM shuts down .......... */
        if( hExecDlg )
            SendMessage( hExecDlg, WM_COMMAND,
                            IDABORT, 0L );
        break;

    case FMEVENT_SELCHANGE:
        break;

    case FMEVENT_INITMENU:
        break;

    /* Handle extension commands as defined in
     * fmx.h ....
     */
    case WINEXEC:	/* Run an application */
        WinExec( pCmdTab->szCmdline, SW_SHOWNORMAL );
        break;

    case SCREENSAVE:    /* screen saver */
        PostMessage( hFMwnd, WM_SYSCOMMAND,
                            SC_SCREENSAVE, 0L );
        break;

    case EXTRELOAD:     /* reload all FM extensions */
        PostMessage( hFMwnd, FM_RELOAD_EXTENSIONS,
                                           0, 0L );
        break;

    case FMRESTORE:   /* Bring FM to top */
        if( IsIconic( hFMwnd ) )
            ShowWindow( hFMwnd, SW_RESTORE );
        else
            BringWindowToTop( hFMwnd );

        break;

    case CMDEXEC:       /* App launcher  */
        GetCurrentFileSel( &file );
        ExecProgram( file.szName );
        break;

    case BINDEXEC:   {	/* Extended App launcher */
        /* Associate (bind) current file selection
         * with a menu application....  */
        static char cmd[260];

        GetCurrentFileSel( &file );

        if( !file.szName[0] )   /* bring up dialog */
            ExecProgram( pCmdTab->szCmdline );
        else    {       /* execute directly */
            memset( (void *) cmd, '\0', sizeof(cmd) );
            lstrcpy( cmd, pCmdTab->szCmdline );
            lstrcat( cmd, file.szName );
            WinExec( cmd, SW_SHOWNORMAL );
            }   /* .... end if */

        }   /* .... end case */

        break;

    case WINHELP:   /* launch Help file */
        WinHelp( hFMwnd, pCmdTab->szCmdline,
                    HELP_INDEX, NULL );
        break;

    case ABOUT:     {
        HWND hwnd = GetFocus();
        MessageBox( hWndPopup,
                   "File Manager Extension\n"
                   "Copyright (c) 1994 T.W. Nelson.\n"
                   "For non-commercial use only.",
                   "About", MB_OK );
        SetFocus( hwnd );
        }

        break;

    case EXECFLOAT:    /* Exec float popup menu */
        if( fEnableFloat )
            PopupFloatMenu();

        break;

    default:
        break;
    }

    return 0L;
}

/* ----- EOF --------------------------------------- */
